/*** -*- Mode: Javascript; tab-width: 2;

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is Collabnet code.
The Initial Developer of the Original Code is Collabnet.

Portions created by Collabnet are
Copyright (C) 2000 Collabnet.  All
Rights Reserved.

Contributor(s): Pete Collins, Doug Turner, Brendan Eich, Warren Harris, Eric Plaster


 File API (The purpose of this file is to make it a little easier to do file IO from js) 

       file.js

 Function List

    // Constructor
    File(aPath)                        // creates the File object and sets the file path
    									// The path may also be a chrome: or file: url.

    open(aMode);                       // open a file handle for reading, writing or appending
    close();                           // closes a file handle
    exists();                          // check to see if a file exists
    read();                            // returns the contents of a file
    readline(aDone);                   // returns the next line in the file.  aDone == true at EOF
    write(aContents, aPermissions);    // permissions are optional 
    eof;                               // check 'end of file' status

    // file attributes
    permissions;                     // returns the files permissions
    dateModified;                    // returns the last modified date in locale string
    size;                            // returns the file size

    // file string manipulation
    path;                            // returns the path
    leaf;                            // leaf is the endmost file string ex: foo.html in /myDir/foo.html
    extension;                       // returns a file extension if there is one
    dirPath;                         // returns the dir part of a path

    // direct manipulation
    nsIFIle()                          // returns an nsIFile obj 

 Instructions:

       First include this js file in your xul file.  Next, create an File object:

          var file = new File("/path/file.ext");

       To see if the file exists, call the exists() member.  This is a good check before going into some
       deep code to try and extract information from a non-existant file.

       To open a file for reading<"r">, writing<"w"> or appending<"a">, 
       just call: 

          file.open("w");

       where in this case you will be creating a new file called '/path/file.ext', with a mode of "w"
       which means you want to write a new file.

       If you want to read from a file, just call:

          file.open("r");
          var theFilesContents    = file.read();

       The file contents will be returned to the caller so you can do something usefull with it.

          file.close(); 

       Calling 'close()' destroys any created objects.  If you forget to use file.close() no probs
       all objects are discarded anyway.

       Warning: these API's are not for religious types

************/
netscape.security.PrivilegeManager.enablePrivilege("UniversalXPConnect");

/****************** Globals **********************/
const FILE_READ                     = 0x01;     // 1
const FILE_WRITE                    = 0x08;     // 8
const FILE_APPEND                   = 0x10;     // 16

const FILE_READ_MODE                = "r";
const FILE_WRITE_MODE               = "w";
const FILE_APPEND_MODE              = "a";

const FILE_DELETE                   = "delete";

const FILE_FILETYPE                 = 0x00;     // 0

const FILE_CHUNK                    = 1024;     // buffer for readline => set to 1k

const FILE_DEFAULT_PERMS            = 0644;

const File_LocalFile     = new Components.Constructor( "@mozilla.org/file/local;1", "nsILocalFile", "initWithPath");

const File_RegistryService = 
	Components.classes["@mozilla.org/chrome/chrome-registry;1"].
	getService(Components.interfaces.nsIChromeRegistry);

const File_IOService = 
	Components.classes["@mozilla.org/network/io-service;1"].
	getService(Components.interfaces.nsIIOService);
	
const File_FileOutputStream = new Components.Constructor("@mozilla.org/network/file-output-stream;1", "nsIFileOutputStream");


const File_ScriptableInputStream = new Components.Constructor("@mozilla.org/scriptableinputstream;1", "nsIScriptableInputStream");

// Possible values for the ioFlags parameter to FileTransportService.createTransport.
// From http://lxr.mozilla.org/seamonkey/source/netwerk/base/public/nsIFileChannel.idl
const FILE_NS_RDONLY               = 0x01;
const FILE_NS_WRONLY               = 0x02;
const FILE_NS_RDWR                 = 0x04;
const FILE_NS_CREATE_FILE          = 0x08;
const FILE_NS_APPEND               = 0x10;
const FILE_NS_TRUNCATE             = 0x20;
const FILE_NS_SYNC                 = 0x40;
const FILE_NS_EXCL                 = 0x80;

const FILE_DIRECTORY_TYPE			= 1;

/****************** Globals **********************/


/****************** File Object Class *********************/

function File(aPath) {
	if(!aPath) throw "File Constructor: No path supplied.";
	try {
		this.mOrigPath		= aPath;
		this.mPath			= null;
		this.mMode 			= null;
		this.mChannel	 	= null;
		this.mInputStream 	= null;
		this.mFileChannel	= null;
		this.mOutputStream 	= null;  
		this.mLineBuffer 	= null;
		this.mFileInst 		= null;
		this.mURI 			= null;
	    this.flushCount = 1; // no of writes before a flush
	    this.writeCount = 0;
    }
    catch(e) {
      throw "File Constructor: Can't create nsILocalFile." + e;
    }
} // constructor
File.prototype.resolve = function() {
	if(!this.mFileInst && !this.mURI) {
		this.mPath = this.mOrigPath;
		if (this.mPath.substr(0, 7) == "chrome:")
			this.mPath = File.chromeToFileURL(this.mPath);
			
		if (this.mPath.substr(0,4) == "jar:"	) {
			this.mFileInst = null; // not a local file.
			this.mURI = File_IOService.newURI(this.mPath,null,null);
		}
		else {
			if (this.mPath.substr(0, 5) != "file:") 
				this.mPath = File.pathToURL(this.mPath);

			this.mURI = File_IOService.newURI(this.mPath,null,null);
			this.mFileInst 		= new File_LocalFile(File.URLToPath(this.mPath));
		}
	}
}

/********************* OPEN *****************************/
File.prototype.open = function(aMode, aPerms) {
	if (!this.mOrigPath) throw "File.open: File not initialized with path.";
	if(this.mMode) throw "File.open: Already open:" + this.mPath;
	this.resolve();
	this.writeCount = 0;

	if(!aMode)
		aMode=FILE_READ_MODE;
    if(!aPerms)
        aPerms=FILE_DEFAULT_PERMS;
	
	if (this.mFileInst) {
		var fileExists = this.exists();

	  	if(fileExists) {
	    	if(this.isDirectory()) throw "File.open: Can't open directory.";
	    }
	}    

   this.mMode=aMode;
	switch(this.mMode) {
	case FILE_WRITE_MODE: 
   		if (!this.mFileInst)
   			throw "File.Open/Write: Can't Open jar file in write mode.";
      	try {
			// Create an output stream.
			this.mOutputStream = new File_FileOutputStream();
      		this.mOutputStream.init(this.mFileInst, FILE_NS_WRONLY | FILE_NS_CREATE_FILE | FILE_NS_TRUNCATE, aPerms, 0);
	 	}
		catch(e) {
			this.close();
			throw "File.open/Write: Unable to openOutputStream.";
		}	
		break;	
    case FILE_APPEND_MODE:
   		if (!this.mFileInst)
   			throw "File.Open/Write: Can't Open jar file in append mode.";
      	try {
      		this.mOutputStream = new File_FileOutputStream();
      		this.mOutputStream.init(this.mFileInst, FILE_NS_WRONLY | FILE_NS_APPEND | FILE_NS_CREATE_FILE , aPerms, 0);
	 	}
		catch(e) {
			this.close();
			throw "File.open/Append: Unable to openOutputStream.\n" + e;
		}
      	break;
    case FILE_READ_MODE:
    	if (this.mFileInst) {
			if(!this.exists()) {
				this.close();
				throw "File.open/Read: File " + this.path + " not found.";
			}
			try {
				// Create a file transport in write mode
				this.mFileChannel = File_IOService.newChannelFromURI(this.mURI);
		        this.mInputStream = new File_ScriptableInputStream();    
		        this.mInputStream.init(this.mFileChannel.open());
				this.mLineBuffer         = new Array();
			} 
			catch (e) {
				this.close();
				throw "File.open/Read: Open failed:\n" + e;
			}
		}
		else {
			try {
				this.mChannel = File_IOService.newChannelFromURI(this.mURI);
				this.mInputStream = new File_ScriptableInputStream();
				this.mInputStream.init(this.mChannel.open());
			} 
			catch (e) {
				this.close();
				throw "File.open/Read: Open failed:\n" + e;
			}
				
		}	
		break;
    default:
		throw 'File.open: Invalid mode "r","w", or "a".';
  }
}

/********************* WRITE ****************************/
File.prototype.write = function(aBuffer) {
	if(this.mMode == FILE_READ_MODE)
		throw "File.write: File is not open in Write or Append mode.";
	if(!this.mOutputStream)
		throw "File.write: File is not open.";

	try {
		this.mOutputStream.write(aBuffer, aBuffer.length);
		this.writeCount++;
		if (this.writeCount >= this.flushCount) {
			this.mOutputStream.flush();
			this.writeCount = 0;
		}	
	}
	catch (e) {  
		throw "File.write: Write error. " + e;
  	}

}

/********************* READ *****************************/
File.prototype.read = function() {
	if(this.mMode != FILE_READ_MODE)
		throw "File.read: File is not open in Read mode.";
	if(!this.mInputStream)
		throw "File.read: File is not open.";

	try {
		if (this.mFileInst)
			return this.mInputStream.read(this.mFileInst.fileSize);
		else {
			return this.mInputStream.read(this.mInputStream.available());
		}
	}
	catch (e) {
		throw this.path + "\nFile.read:\n" + e;;
	}
}

/********************* READLINE *****************************/
File.prototype.readline = function() {
	try {
		if(this.mMode != FILE_READ_MODE)
			throw "File.readline: File is not open in Read mode.";
		if(!this.mInputStream)
			throw "File.readline: File is not open.";

		if(this.mLineBuffer.length < 2) {
			var buf = this.mInputStream.read(FILE_CHUNK);
			if(buf) {
				if(this.mLineBuffer.length == 1) {
			  		var tmp = this.mLineBuffer.shift();
			  		buf = tmp+buf;
				}
				this.mLineBuffer = buf.split(/[\n\r]/);
			}
		}
		return this.mLineBuffer.shift();
	}
	catch (e){ 
		throw "File.readline: Read error:" + e;
	}
}

/********************* EOF *****************************/
File.prototype.__defineGetter__('eof', function() {

  if(!this.mMode != FILE_READ)
  	throw "File.eof: file is not open in Read mode.";

  if((this.mLineBuffer.length > 0) || (this.mInputStream.available() > 0)) 
    return false;
  else 
    return true;
  
})
/********************* PATH *****************************/
File.prototype.__defineGetter__('path', function () {
	if (!this.mFileInst && !this.mURI)
		this.resolve();
	if (this.mFileInst)
		return this.mFileInst.path;
	return this.mURI.spec;	
})

/********************* PARENT *****************************/
File.prototype.__defineGetter__('parent', function () {
	if (!this.mFileInst && !this.mURI)
		this.resolve();
	if (!this.mFileInst) throw "File.parent: Only allowed for local files: " + this.mOrigPath;
	return new File(this.mFileInst.parent.path);
})

/********************* PATH *****************************/
File.prototype.__defineGetter__('URL', function () {
	return File.pathToURL(this.path);
})

/********************* EXISTS ***************************/
File.prototype.exists = function () {
	if (!this.mFileInst && !this.mURI)
  		this.resolve();
	if (!this.mFileInst) throw "File.exists: Only allowed for local files: " + this.mOrigPath;
  		
	try {
		return this.mFileInst.exists();
	}
	catch(e) { 
		throw "File.exists: Unknown error:" + e;
	}
}


/********************* LEAF *****************************/
File.prototype.__defineGetter__('leaf', function () {
	if (!this.mFileInst && !this.mURI)
  		this.resolve();
	if (!this.mFileInst) throw "File.leaf: Only allowed for local files: " + this.mOrigPath;

	try  {
		return this.mFileInst.leafName;
	}
	catch(e) {
		throw "File.leaf: Error: " + e;
	}
})


/********************* PERMISSIONS **********************/
File.prototype.__defineGetter__('permissions', function () {
	if(this.exists()) {
		return this.mFileInst.permissions.toString(8);
	}
	else
	  	throw "File.permissions: File " + this.path + " does not exist.";
})

/********************* MODIFIED *************************/
File.prototype.__defineGetter__('dateModified', function () {
  if(this.exists()) {
  	return new Date(this.mFileInst.lastModificationDate).toLocaleString();
  }
  else
  	throw "File.dateModified: File " + this.path + " does not exist.";
})

/********************* SIZE *****************************/
File.prototype.__defineGetter__('size', function ()  {
  if(this.exists()) {
	  try { 
		return this.mFileInst.fileSize;
	  }
	  catch(e) { 
			throw "File.size: Error: " + e;
	  }
  }
  else
  	throw "File.size: File " + this.path + " does not exist.";
})

/********************* EXTENSION ************************/
File.prototype.__defineGetter__('extension', function () {
	if(this.exists()) {
		try { 
			var leafName  = this.mFileInst.leafName; 
			var dotIndex  = leafName.lastIndexOf('.'); 
			return (dotIndex >= 0) ? leafName.substring(dotIndex+1) : ""; 
		}
		catch(e) {
			throw "File.extension: Error: " + e;
		}
	}
	else
	  	throw "File.extension: File " + this.path + " does not exist.";

})

/********************* DIRPATH **************************/
File.prototype.__defineGetter__('dirPath', function () {
	if(this.exists()) {
		try { 
			if(this.mFileInst.isFile())
			  return this.mFileInst.parent.path;

			if(this.mFileInst.isDirectory())
			  return this.mFileInst.path;
		}
		catch (e) { 
			throw "File.dirPath: Error: " + e;
		}
	}
	else
		throw "File.dirPath: File " + this.path + " does not exist.";
	
})

/********************* nsIFILE **************************/
File.prototype.__defineGetter__('nsILocalFile', function () {
	if (!this.mFileInst && !this.mURI)
		this.resolve();

	return this.mFileInst;
})

/********************* CLOSE ****************************/
File.prototype.close = function() {
  /***************** Destroy Instances *********************/
	if(this.mFileChannel)   this.mFileChannel = null;
	if(this.mChannel)   this.mChannel = null;  
	if(this.mInputStream) this.mInputStream = null;
	if(this.mOutputStream) this.mOutputStream = null;
	if(this.mMode)         this.mMode=null;
}

File.prototype.isFile = function() {
	if (!this.mFileInst && !this.mURI)
		this.resolve();

	try {
		if (mFileInst)
			return this.mFileInst.isFile();
		else
			return false;
	}
	catch (e) { 
		throw "File.isFile: Error: " + e;
	}
}

File.prototype.isDirectory = function() {
	if (!this.mFileInst && !this.mURI)
		this.resolve();

	try { 
		if (this.mFileInst)
			return this.mFileInst.isDirectory();
		else
			return false;
	}
	catch (e) { 
		throw "File.isDirectory: Error: " + e;
	}
}

File.prototype.isWritable = function() {
	if (!this.mFileInst && !this.mURI)
		this.resolve();

	try { 
		return this.mFileInst.isWritable();
	}
	catch (e) { 
		throw "File.isWritable: Error: " + e;
	}
}

File.prototype.createDir = function(permissions) {
	if (!this.mFileInst && !this.mURI)
		this.resolve();

	try { 
		if (!permissions) permissions = FILE_DEFAULT_PERMS;
		this.mFileInst.create(FILE_DIRECTORY_TYPE, permissions);
		return this;
	}
	catch (e) { 
		throw "File.create: Error: " + e;
	}
}

File.prototype.clone = function() {
	try { 
		return new File(this.path);
	}
	catch (e) { 
		throw "File.clone: Error: " + e;
	}
}

/**TIP: To get correct native file separator's append one level at a time. **/
File.prototype.append = function(pathToAppend) {
	if (!this.mFileInst && !this.mURI)
		this.resolve();

	try { 
		this.mFileInst.append(pathToAppend);
		// and update mURI to reflect change.
		this.mURI = File_IOService.newFileURI(this.mFileInst);
		return this;
	}
	catch (e) { 
		throw "File.clone: Error: " + e;
	}
}

/**
 * Convenience function for converting chrome to file URL 
**/
File.chromeToFileURL= function(chromeURL)  {
	var  uriObj = File_IOService.newURI(chromeURL,null,null);  
    return File_RegistryService.convertChromeURL(uriObj).toString();
}
/**
 * Convenience function for converting file URL to native path. 
**/
File.URLToPath = function(fileURL) {
	var path;
	if (fileURL.substr(0,8) == "file:///")
		path = fileURL.substr(8);
	else {	
		if (fileURL.substr(0,7) == "file://")
			path = fileURL.substr(7);
		else
			path = fileURL;
	}		 
	path = path.replace(/%20/g," "); // Replace space references.
	if (File.isWindows()) { 
		path = path.replace(/\//g,"\\"); // Replace forward by retarded slashes.
		path = path.replace(/\|/g,":"); // Replace drive separator.
	}	
	return path;
}
/**
 * Convenience function for converting native path to file URL. 
**/
File.pathToURL = function(path) {
	if (path.substr(0,7) == "file://")
		return path;
	if (path.substr(0,9) == "chrome://")
		return path;

	var URL;
	URL = path.replace(/\s/g,"%20"); // Replace space references.
	if (File.isWindows()) { 
		URL = URL.replace(/\\/g,"/"); // Replace retarded slashes by forward slashes.
		URL = URL.replace(/:/g,"|"); // Replace drive separator.
	}	
	return "file://" + URL;
}

File.dirPath = function(path) {
	if (!path)
		return null;
	var sep;
	if (path.substr(0,7) == "chrome:" || path.substr(0,5) == "file:")
		sep = "/";
	else
		sep = File.isWindows()? "\\" : "/"; 
	var i = path.lastIndexOf(sep);
	if (i > -1)
		return path.substring(0,i);
	else
		return path;
}

File.isWindows = function() {
	return (navigator.appVersion.indexOf("Win") != -1); 
}
